import time
import logging
import asyncio

from manager import ops

log = logging.getLogger(__name__)


class WebAPIError(Exception):
    def __init__(self, code, msg):
        super().__init__(msg)
        self.code = code


class WebAPI:
    """
    Backend-independent functions exposed via REST or WebSocket APIs
    """
    def __init__(self, manager):
        self.manager = manager

    async def __call__(self, function=None, **kw):
        """
        Call a web API function by name and keyword arguments
        """
        log.debug("API call %s %r", function, kw)
        if function is None:
            log.debug("API call %s %s: function is missing", function, kw)
            return None
        f = getattr(self, "do_" + function)
        if f is None:
            log.debug("API call %s %s: function not found", function, kw)
            return None
        if asyncio.iscoroutinefunction(f):
            res = await f(**kw)
        else:
            res = f(**kw)
        log.debug("API call %s %r result %r", function, kw, res)
        res["time"] = time.time()
        return res

    def do_ping(self, **kw):
        return {
            "pong": True,
        }

    async def do_async_ping(self, **kw):
        return {
            "pong": True,
        }

    def do_list_machines(self):
        return {
            "machines": [
                m.to_jsonable()
                for m in self.manager.store.machines.values()
                ],
        }

    def do_get_machine(self, mac):
        res = self.manager.store.machines.get(mac)
        if res is None:
            raise WebAPIError(404, "machine not found")
        return {
            "machine": res.to_jsonable(),
        }

    async def do_perform_operation(self, op):
        op = ops.Operation.from_json(op)
        await self.manager.execute(op)
        return {}

    def do_list_playbooks(self):
        return {
            "playbooks": list(self.manager.ansible.list_playbooks()),
        }

    def do_get_playbook_results(self, id):
        res = self.manager.ansible.playbook_log.load(id)
        if res is None:
            raise WebAPIError(404, "machine not found")
        return {
            "playbook": res,
        }
