import asyncio
import tempfile
import os

from .common import gen_test, InotifyTestCase

from manager.sources.inotifywait import Inotifywait


class TestInotifywatch(InotifyTestCase):
    """
    Test the Inotifywatch helper
    """
    @gen_test
    def test_start_stop(self):
        with tempfile.TemporaryDirectory() as path:
            fname = os.path.join(path, "testfile")
            watch = Inotifywait(fname)

            @asyncio.coroutine
            def do_test():
                self.assertEqual((yield from watch.next_event()), "START")
                watch.stop()
                self.assertEqual((yield from watch.next_event()), "END")

            yield from asyncio.gather(watch.run(), do_test())

    @gen_test
    def test_write(self):
        with tempfile.TemporaryDirectory() as path:
            fname = os.path.join(path, "testfile")
            watch = Inotifywait(fname)

            @asyncio.coroutine
            def do_test():
                self.assertEqual((yield from watch.next_event()), "START")
                # Do a normal write on the file
                with open(fname, "at") as fd:
                    print("UPDATE", file=fd)
                    # No changes detected while the file is still open
                    self.assertFalse(watch.has_events())
                self.assertEqual((yield from watch.next_event()), "MODIFIED")
                watch.stop()
                self.assertEqual((yield from watch.next_event()), "END")

            yield from asyncio.gather(watch.run(), do_test())

    @gen_test
    def test_atomic_write(self):
        with tempfile.TemporaryDirectory() as path:
            fname = os.path.join(path, "testfile")
            watch = Inotifywait(fname)

            @asyncio.coroutine
            def do_test():
                self.assertEqual((yield from watch.next_event()), "START")
                # Do an atomic update via move
                with open(fname + ".tmp", "at") as fd:
                    print("UPDATE", file=fd)
                # No changes detected while only the tempfile has been written
                self.assertFalse(watch.has_events())
                os.rename(fname + ".tmp", fname)
                # The move is detected
                self.assertEqual((yield from watch.next_event()), "MODIFIED")
                watch.stop()
                self.assertEqual((yield from watch.next_event()), "END")

            yield from asyncio.gather(watch.run(), do_test())
