from urllib.parse import urlencode
import json
import tempfile
from unittest import mock

from tornado.testing import AsyncHTTPTestCase, gen_test
from tornado import gen

from manager.web.server import Application
from manager.config import Config

from .common import AsyncTestCase


class TestManagerApplication(AsyncTestCase, AsyncHTTPTestCase):
    def get_app(self, *args, **kw):
        config = Config()
        op_log = tempfile.NamedTemporaryFile()
        config.operation_log_file = op_log.name
        return Application(config)

    def test_root(self):
        resp = self.fetch('/')
        self.assertEqual(resp.code, 200)


class TestWebAPI(AsyncTestCase, AsyncHTTPTestCase):
    def get_app(self, *args, **kw):
        config = Config()
        op_log = tempfile.NamedTemporaryFile()
        config.operation_log_file = op_log.name
        return Application(config)

    @gen.coroutine
    def api_get(self, name, expected_code=200, data=None):
        url = "/api/1.0/{}".format(name)
        if data:
            url += "?" + urlencode(data)
        response = yield self.http_client.fetch(
            self.get_url(url),
            raise_error=False
            )
        content_type = response.headers.get("Content-Type")
        self.assertIn("application/json", content_type)
        self.assertEqual(response.code, expected_code)
        return json.loads(response.body.decode())

    @gen.coroutine
    def api_post(self, name, expected_code=200, data=None):
        url = "/api/1.0/{}".format(name)
        if data is None:
            data = {}
        body = json.dumps(data)
        response = yield self.http_client.fetch(
            self.get_url(url),
            method="POST",
            body=body,
            raise_error=False
            )
        content_type = response.headers.get("Content-Type")
        if "application/json" not in content_type:
            print("*** Response with unexpected content type", content_type)
            print(response.body.decode())
        self.assertIn("application/json", content_type)
        self.assertEqual(response.code, expected_code)
        return json.loads(response.body.decode())

    @gen_test
    def test_ping(self):
        with mock.patch("time.time", return_value=100):
            res = yield self.api_get("ping")
        self.assertEqual(res, {
            "pong": True,
            "time": 100,
        })

    @gen_test
    def test_async_ping(self):
        with mock.patch("time.time", return_value=100):
            res = yield self.api_get("async_ping")
        self.assertEqual(res, {
            "pong": True,
            "time": 100,
        })

    @gen_test
    def test_machines(self):
        with mock.patch("time.time", return_value=100):
            res = yield self.api_get("machines")
        self.assertEqual(res, {
            "machines": [],
            "time": 100,
        })

    @gen_test
    def test_machine(self):
        with mock.patch("time.time", return_value=100):
            res = yield self.api_get(
                "machine/11:22:33:44:55:66",
                expected_code=404
                )
        self.assertEqual(res, {
            "code": 404,
            "error": True,
            "message": "machine not found",
            "time": 100,
        })

    @gen_test
    def test_operation(self):
        with mock.patch("time.time", return_value=100):
            res = yield self.api_post(
                "operation",
                data={"op": {"name": "Noop"}}
                )
        self.assertEqual(res, {
            "time": 100,
        })
