import asyncio
import time
import json
import logging

from . import sources, stores, events, playbook


class OperationLog:
    """
    Logger for Operation objects that get executed by the Manager
    """
    def __init__(self, pathname):
        self.pathname = pathname
        try:
            self.fd = open(self.pathname, "at")
        except FileNotFoundError:
            self.fd = None

    def close(self):
        if self.fd is None:
            return
        self.fd.close()

    def log(self, op):
        if self.fd is None:
            return
        op_data = op.to_jsonable()
        op_data["ts"] = time.time()
        json.dump(op_data, self.fd)
        self.fd.write("\n")


class Manager:
    def __init__(self, config):
        self.log = logging.getLogger(self.__class__.__name__)
        self.event_hub = events.Hub()
        self.config = config
        self.sources = []
        self.load_data_sources()
        self.store = stores.MachineStore(self.event_hub, config=self.config)
        self.operation_log = OperationLog(self.config.operation_log_file)
        self.ansible = playbook.Ansible(self.event_hub, self.config)
        self.users = None

    def load_data_sources(self):
        """
        Try to load all known data sources.

        Check for viability using .is_viable() to prevend loading data
        sources that are missing dependencies.
        """
        # sources that require no params
        for source in [
                sources.DhcpdDataSource,
                sources.ArpwatchDataSource,
                sources.ChaosDataSource,
                ]:
            if not source.is_viable(self.config):
                self.log.warn(
                    "%s: data source will not be started",
                    source.__name__
                    )
                continue
            self.sources.append(
                source(event_hub=self.event_hub, config=self.config)
                )

    async def load_user_database(self):
        """
        Try to load a user database
        """
        from .users import ldap
        from .users import local
        from .users import mock
        # sources that require no params
        for db in [
                ldap.LDAP,
                local.Local,
                mock.Mock,
                ]:
            if not await db.is_viable(self.config):
                self.log.warn(
                    "%s: user database will not be used",
                    db.__name__
                    )
                continue
            return db(self.config)
        raise RuntimeError("Cannot instantiate a valid user database")

    async def run(self):
        self.users = await self.load_user_database()
        await self.store.load()
        coros = []
        for ds in self.sources:
            coros.append(ds.run())
        await asyncio.gather(*coros)

    async def shutdown(self):
        self.log.info("shutdown requested")
        coros = []
        for ds in self.sources:
            coros.append(ds.shutdown())
        await asyncio.gather(*coros)
        self.operation_log.close()

    async def execute(self, op):
        """
        Execute an operation
        """
        self.operation_log.log(op)
        await op.execute(self)
