import asyncio
import shutil

from unittest import TestCase, SkipTest
from manager import events, compat
import functools


class StopTest(events.Event):
    pass


class GatherEvents:
    def __init__(self, event_hub, event_filter=None):
        self.gathered = []
        loop = asyncio.get_event_loop()
        self.received_first = compat.create_future(loop)
        event_hub.subscribe(self.gather_events, event_filter)

    async def gather_events(self, evt):
        if not self.received_first.done():
            self.received_first.set_result(True)
        self.gathered.append(evt)


class TestDataSourceMixin:
    """
    Start a data source on setUp and stop it when StopTest is posted on the
    event hub
    """
    def setUp(self, *args, **kw):
        super().setUp(*args, **kw)
        self.event_hub = events.Hub()
        self.event_hub.subscribe(self.stop_test, [StopTest])
        self.ds = self.make_datasource()

    async def stop_test(self, evt):
        await self.ds.shutdown()


# version of tornado.testing.TestCase that works with asyncio coroutines
# instead of django gen.coroutines. Use tornado's version when we can drop
# Jessie support.
class AsyncTestCase(TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.loop.set_debug(True)
        super().setUp()

    def tearDown(self):
        super().tearDown()
        try:
            pending = asyncio.Task.all_tasks()
            try:
                self.loop.run_until_complete(asyncio.gather(*pending))
            except asyncio.CancelledError:
                pass
        finally:
            asyncio.set_event_loop(None)
            self.loop.close()

    def get_new_ioloop(self):
        """
        Integrate the asyncio ioloop with the pre-5.0 tornado AsyncTestCase
        """
        from tornado.platform.asyncio import AsyncIOMainLoop
        return AsyncIOMainLoop()


# version of tornado.testing.get_test that works with asyncio coroutines
# instead of django gen.coroutines. Use tornado's version when we can drop
# Jessie support.
def gen_test(testfunc):
    testfunc = asyncio.coroutine(testfunc)

    @functools.wraps(testfunc)
    def wrap(self):
        self.loop.run_until_complete(testfunc(self))

    return wrap


class FussTestNetwork:
    def __init__(self):
        self.require_test_machines()

    def require_test_machines(self):
        """
        Raise SkipTest unless the fuss-test-network VMs are up and running
        """
        import socket
        try:
            conn = socket.create_connection(("192.168.123.2", 22), timeout=0.1)
        except (ConnectionRefusedError, OSError):
            raise SkipTest("test network is not set up")
        else:
            conn.close()

    def write_inventory(self, fd, group_name="test"):
        print("[{}]".format(group_name), file=fd)
        for name, ip in (
                    ("test1", "192.168.123.2"),
                    ("test2", "192.168.123.3"),
                ):
            print(('{name} ansible_host={ip} ansible_user=root'
                   ' ansible_ssh_extra_args="-oUserKnownHostsFile=/dev/null -oStrictHostKeyChecking=no"'
                   ).format(name=name, ip=ip), file=fd)


class InotifyTestCase(AsyncTestCase):
    def setUp(self):
        path_inotifywait = shutil.which("inotifywait")
        if not path_inotifywait:
            raise SkipTest("inotify is not available")
        super().setUp()
